class JournalError(Exception):
    """Base error for all accumulator functionality."""


class ResultNameMappingError(JournalError):
    """Error mapping results."""


class FormatterError(JournalError):
    """Unable to format log message."""


class EncoderError(JournalError):
    """Error dictionary encoding object."""



import copy
from inspect import getcallargs
from typing import Iterable, Mapping, Callable, Optional, List, Union, Tuple, Any

from callable_journal.exceptions import ResultNameMappingError


class ParamArgMapper:
    """
    Map the argument values and parameters.
    """

    COPY_ALL = ["*"]
    IGNORE = "_"

    @classmethod
    def map_args(
        cls,
        callable: Callable,
        args: Iterable,
        kwargs: Mapping,
        copy_args: Optional[List[str]] = None,
    ) -> Mapping:
        """
        Map the argument values onto the parameters.  Remove self or cls if they
        exist in the parameters.  Make copies of any arguments that are mutable
        and my be changed by the wrapped callable.

        :param callable: Callable to map the arguments to the parameter signature.
        :param args: List of positional arguments.
        :param kwargs: Dictionary of keyword arguments.
        :param copy_args: List of argument names to make deep copies of.  ParamArgMap.copy_all
            to map all arguments.
        :return: Dictionary of the mapped arguments.
        """
        arg_map = getcallargs(callable, *args, **kwargs)
        for k in ["self", "cls"]:
            try:
                del arg_map[k]
            except KeyError:
                pass

        if copy_args:
            if "*" in copy_args:
                copy_args = arg_map.keys()
            for k in copy_args:
                arg_map[k] = copy.deepcopy(arg_map[k])
        return arg_map

    @classmethod
    def map_results(
        cls,
        results: Union[Any, Tuple[Any]],
        result_names: Optional[Union[str, List[str]]] = None,
    ) -> Mapping:
        """
        Map results based on list of names.  Result names "_" are ignored and not
        mapped into the results dictionary.

        :param results:  Result values to map.
        :param result_names: Name or list of result names.
        :return: Dictionary of named results.
        """
        if not result_names:
            return results

        # Make sure the results and names are lists.
        results = results if isinstance(results, (list, tuple)) else [results]
        result_names = (
            result_names if isinstance(result_names, (list, tuple)) else [result_names]
        )
        if len(result_names) != len(results):
            raise ResultNameMappingError(
                f"Results names: {len(result_names)}, Results: {len(results)}."
            )

        # Map everything together and throw out the "_".
        mapped_results = {
            n: r for n, r in zip(result_names, results) if n != cls.IGNORE
        }
        return mapped_results



import functools
import logging.config
import traceback
from pathlib import Path
from typing import Any, List, Optional, Mapping, Union, Dict

import yaml
from pydantic.main import BaseModel
from toolz import curry

from callable_journal.param_arg_mapper import ParamArgMapper

logger = logging.getLogger("journal")


class ExceptionMsg(BaseModel):
    type: str
    msg: str
    file: str
    line: str

    @classmethod
    def from_exception(cls, exc: Exception) -> "ExceptionMsg":
        """
        Extract the subset of the exception information that will be saved in the exceptions list.

        :param exc:  Exception to extract
        """
        exc_type = exc.__class__.__name__
        exc_msg = str(exc)
        exc_frame = traceback.extract_tb(exc.__traceback__)[-1]
        exc_file = exc_frame.filename
        exc_line = exc_frame.lineno
        return ExceptionMsg(type=exc_type, msg=exc_msg, file=exc_file, line=exc_line)


class SimpleCtx(BaseModel):
    name: str
    version: str


class JournalContent(BaseModel):
    context: Dict[str, SimpleCtx]
    objective: str
    arguments: Mapping[str, Any]
    results: Optional[Mapping[str, Any]] = None
    exception: Optional[ExceptionMsg] = None


class JournalCtx(BaseModel):
    context: Dict[str, SimpleCtx]


ctx = None


def journal_init(journal_ctx: JournalCtx, logging_cfg_fpath: Path):
    global ctx
    ctx = journal_ctx

    with logging_cfg_fpath.open("rt") as fp:
        logging_cfg = yaml.safe_load(fp)

    logging.config.dictConfig(logging_cfg)


@curry
def journal(
    func,
    *,
    objective: str,
    results_names: Optional[Union[str, List[str]]] = None,
    copy_args: Optional[List[str]] = None
):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        args_data = ParamArgMapper.map_args(func, args, kwargs, copy_args=copy_args)
        msg = JournalContent(
            context=ctx.context,
            objective=objective,
            arguments=args_data,
        )
        try:
            results = func(*args, **kwargs)
            results_data = ParamArgMapper.map_results(
                results, result_names=results_names
            )
            msg.results = results_data
            logger.info(msg="", extra={"journal_content": msg})
            return results
        except Exception as exc:
            msg.exception = ExceptionMsg.from_exception(exc)
            logger.exception(msg="", extra={"journal_content": msg})
            raise

    return wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['callable_journal']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.6.1,<2.0.0', 'pyyaml>=5.3.1,<6.0.0', 'toolz>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'callable-journal',
    'version': '0.0.2',
    'description': 'Log message generator for callables argument and return values.',
    'long_description': '# callable-journal\n[![CI](https://github.com/nathan5280/callable-journal/workflows/Test/badge.svg)](https://github.com/nathan5280/callable-journal/actions)\n[![coverage](https://codecov.io/gh/nathan5280/callable-journal/master/graph/badge.svg)](https://codecov.io/gh/nathan5280/callable-journal)\n[![pypi](https://img.shields.io/pypi/v/callable-journal.svg)](https://pypi.python.org/pypi/callable-journal)\n[![versions](https://img.shields.io/pypi/pyversions/callable-journal.svg)](https://github.com/nathan5280/callable-journal)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://github.com/nathan5280/callable-journal/blob/master/LICENSE)\n\nLog message generator for callables argument and return values.',
    'author': 'Nate Atkins',
    'author_email': 'atkinsnw@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nathan5280/ndl-tools',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Dictionary encoders to convert objects to primitive dictionaries."""
import datetime
from abc import abstractmethod, ABC
from pathlib import Path
from typing import Mapping, Iterable, Any, Union

PRIMITIVE = Union[bool, float, int, str, Iterable, Mapping]


class DictEncoder(ABC):
    """Base dictionary encoder with logic to traverse str, Iterable and Mapping."""

    @classmethod
    def _mapping_encode(cls, map_: Mapping[str, Any]) -> Mapping:
        """
        Encode a object that implements the Mapping protocol.
        """
        output = dict()
        for key, value in map_.items():
            output[key] = cls.encode(value)
        return output

    @classmethod
    @abstractmethod
    def encode(cls, obj: Any) -> PRIMITIVE:
        """
        Encode common container primitives

        :param obj: Object to encode
        :return: Primitive encoding of object.
        """
        if isinstance(obj, str):
            return obj
        if isinstance(obj, Mapping):
            return cls._mapping_encode(obj)
        if isinstance(obj, Iterable):
            return [cls.encode(v) for v in obj]
        return obj


class ObjectDictEncoder(DictEncoder):
    """
    Encoder for additional common object types and protocols that have a dict() method.
    """

    @classmethod
    def encode(cls, obj):
        try:
            # Try converting with dict() method first.
            return cls.encode(obj.dict())
        except AttributeError as exc:
            pass
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        if isinstance(obj, Path):
            return str(obj)
        return super().encode(obj)



"""Logging formatters to convert Accumulator data to log message format."""
import importlib
import json
import logging
from enum import Enum
from logging import Formatter
from typing import Dict, Type, Optional

from .encoders import ObjectDictEncoder, DictEncoder

logger = logging.getLogger(__name__)

FORMAT_VERSION = "0.2.0"


class FormatMode(Enum):
    """Configure the formatting mode."""

    json = "JSON"
    stringy = "STRINGY"


class JournalFormatter(Formatter):
    def __init__(
        self,
        tag: str,
        format_mode: str,
        *args,
        encoder: Optional[DictEncoder] = None,
        **kwargs
    ):
        self.tag = tag
        format_mode = FormatMode(format_mode)
        if format_mode == FormatMode.json:
            self.formatter = self.format_json
        else:
            self.formatter = self.format_stringy
        self.encoder = encoder if encoder else ObjectDictEncoder
        super().__init__(*args, **kwargs)

    def format_json(self, record) -> Dict:
        msg = {
            "tag": self.tag,
            "version": FORMAT_VERSION,
            **self.encoder.encode(record.journal_content),
        }
        return msg

    def format_stringy(self, record) -> Dict:
        msg = self.format_json(record)
        # Stringy the arguments, results and exceptions.
        for field in ("arguments", "results", "exception"):
            msg[field] = json.dumps(msg[field])
        return msg

    def format(self, record) -> str:
        return json.dumps(self.formatter(record))
